/* 
   Parallel Biofilm Simulation for 2 Processors
   By Shay Ellison
   7/3/9
   Program to accompany module 
   "Biofilms:  United They Stand, Divided They Colonize"
   by Angela B. Shiflet and George W. Shiflet
*/

#include <stdlib.h>
#include <stdio.h>

#define SIMPLE_SPRNG
#define USE_MPI
#include <mpi.h>
#include <sprng.h>

#define ROOT 0

#define TOP_ROW 0
#define BOT_ROW 1
#define MES_TAG 2
#define CHANGED_ROOT 3
#define CHANGED_OTHER 4

#define max(a, b) ((a) > (b)) ? (a) : (b)

#define MAX_NUTRIENT 1.0
#define CONSUMED 0.1

#define EMPTY 0
#define BACTERIUM 1
#define DEAD 2
#define BORDER 3
#define TAKEN 10

/* struct for storing a grid location and whether the location is on another processor */
typedef struct location {
	int row, col, otherProc;
} GridLocation;

/* return a random number between [0, 1), for the parallel algorithm sprng is used */
float randomReal(void);

/* return a random integer between [0, UpperBound) */
int random_int(int UpperBound);

void printNutrientMatrix(float **matrix, int rows, int cols);

void printBacteriaMatrix(int **matrix, int rows, int cols);

/* Init all cells in nutrient grid, including boundaries*/
void initNutrientGrid(int rows, int cols, float **nutrientGrid);

/* */
void applyDiffusion(float **oldNutrientGrid, float **newNutrientGrid, float diffusionRate, 
		int rows, int cols);

/* Return new nutrition value of cell */
float diffusion(float rate, float site, float n, float ne, float e, float se, float s, float sw, 
		float w, float nw);

/* Init all cells in bacteria grid, including boundaries */
void initBacteriaGrid(int rows, int cols, int **bacteriaGrid, float probInitBacteria);

/* probGrow helper function, computes sum of nutrients of all living cells */
float sumNutritionBacteria(int **bacteriaGrid, float **nutrientGrid, int rows, int cols);

/* Return p/E nutrition, i.e. the probability that a cell will divide */
float probGrow(int **bacteriaGrid, float **nutrientGrid, float p, int rows, int cols);

/* return location of chosen neighbor to grow into for two processors */
GridLocation parallel2PickNeighbor(int row, int col, int rows, int cols, int n, int e, int s, int w);

/* */
void parallel2Grow(int **oldBacteriaGrid, int **newBacteriaGrid, float **nutrientGrid, float p, int rows, int cols, int procRank);

/* */
void processParallelGrowingCells(int **bacteriaGrid, int *locations);

/* */
void consume(float **oldNutrientGrid, float **newNutrientGrid, int **bacteriaGrid, int rows, int cols);

int main(int argc, char *argv[]) {
	int rows, cols, rowsPerProcessor, colsPerProcessor, i, j, t, timeSteps;
	int **oldBacteriaGrid, **newBacteriaGrid;
	float **oldNutrientGrid, **newNutrientGrid, probInitBacteria, diffusionRate, p;
	int rowsProcBoundaries, colsProcBoundaries;
	int procRank, numProc;
	
	/* Information about Biofilm simulation */
	rows = cols = 5;
	timeSteps = 10;
	probInitBacteria = 0.5;
	diffusionRate = 0.1;
	p = 1;
	
	/* MPI_COMM_WORLD is our chosen communication network which includes all processors running, in this case 2
           * we put in under the name comm of type MPI_Comm for ease of use later on */
	MPI_Comm comm = MPI_COMM_WORLD;
	MPI_Status status;
	
	/* MPI Init and Set-up */
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(comm, &procRank);
	MPI_Comm_size(comm, &numProc);
	
	/* Domain Decomposition -
           * Here we are dividing the domain into rows / numProc rows with cols number of columns
	 * So if our original matrix has 10 rows, each processor will have a domain of 5 rows */
	rowsPerProcessor = rows / numProc;
	colsPerProcessor = cols;
	
	/* */
	rowsProcBoundaries = rowsPerProcessor + 2;
	colsProcBoundaries = colsPerProcessor + 2;
	
	oldNutrientGrid = malloc(sizeof(float*) * rowsProcBoundaries); 
	newNutrientGrid = malloc(sizeof(float*) * rowsProcBoundaries); 
	oldBacteriaGrid = malloc(sizeof(int*) * rowsProcBoundaries); 
	newBacteriaGrid = malloc(sizeof(int*) * rowsProcBoundaries); 
	for (i = 0; i < rowsProcBoundaries; i++) {
		oldNutrientGrid[i] = malloc(sizeof(float) * colsProcBoundaries);
		newNutrientGrid[i] = malloc(sizeof(float) * colsProcBoundaries);
		oldBacteriaGrid[i] = malloc(sizeof(int) * colsProcBoundaries);
		newBacteriaGrid[i] = malloc(sizeof(int) * colsProcBoundaries);
	}	
	initNutrientGrid(rowsPerProcessor, colsPerProcessor, oldNutrientGrid);
	initBacteriaGrid(rowsPerProcessor, colsPerProcessor, oldBacteriaGrid, probInitBacteria);
	
	/* Simulation */
	for (t = 0; t < timeSteps; t++) {
		/* ghost cells from neighbors */
		if (procRank == ROOT) {
			/* top and bottom rows */
			/* ------------nutrient grid--------------- */
			/* send bottom row to processor 1's top ghost row */
			MPI_Send(&oldNutrientGrid[rowsPerProcessor][1], colsPerProcessor, MPI_FLOAT, 1, BOT_ROW, comm); 
			/* recv processor 1's top, place in bottom ghost row */
			MPI_Recv(&oldNutrientGrid[rowsPerProcessor + 1][1], colsPerProcessor, MPI_FLOAT, 1, TOP_ROW, comm, &status);
			/* send top row to processor 1's bottom ghost row */
			MPI_Send(&oldNutrientGrid[1][1], colsPerProcessor, MPI_FLOAT, 1, TOP_ROW, comm);
			/* receive processor 1's bottom row and place in top ghost row */
			MPI_Recv(&oldNutrientGrid[0][1], colsPerProcessor, MPI_FLOAT, ROOT, BOT_ROW, comm, &status);
			/* -----------bacteria grid---------------- */
			MPI_Send(&oldBacteriaGrid[rowsPerProcessor][1], colsPerProcessor, MPI_FLOAT, 1, BOT_ROW, comm); 
			MPI_Recv(&oldBacteriaGrid[rowsPerProcessor + 1][1], colsPerProcessor, MPI_FLOAT, 1, TOP_ROW, comm, &status);
			MPI_Send(&oldBacteriaGrid[1][1], colsPerProcessor, MPI_FLOAT, 1, TOP_ROW, comm);
			MPI_Recv(&oldBacteriaGrid[0][1], colsPerProcessor, MPI_FLOAT, ROOT, BOT_ROW, comm, &status);
		} else {
			/* top and bottom rows */
			/* ------------nutrient grid--------------- */
			/* receive ROOT's bottom row, place in top ghost row */
			MPI_Recv(&oldNutrientGrid[0][1], colsPerProcessor, MPI_FLOAT, ROOT, BOT_ROW, comm, &status);
			/* send top row to ROOT's bottom ghost row */
			MPI_Send(&oldNutrientGrid[1][1], colsPerProcessor, MPI_FLOAT, ROOT, TOP_ROW, comm);
			/* receive ROOT's top row, place in bottom ghost row */
			MPI_Recv(&oldNutrientGrid[rowsPerProcessor + 1][1], colsPerProcessor, MPI_FLOAT, ROOT, TOP_ROW, comm, &status);
			/* send bottom row to ROOT's top ghost row */
			MPI_Send(&oldNutrientGrid[rowsPerProcessor][1], colsPerProcessor, MPI_FLOAT, ROOT, BOT_ROW, comm); 
			/* -----------bacteria grid---------------- */
			MPI_Recv(&oldBacteriaGrid[0][1], colsPerProcessor, MPI_FLOAT, ROOT, BOT_ROW, comm, &status);
			MPI_Send(&oldBacteriaGrid[1][1], colsPerProcessor, MPI_FLOAT, ROOT, TOP_ROW, comm);
			MPI_Recv(&oldBacteriaGrid[rowsPerProcessor + 1][1], colsPerProcessor, MPI_FLOAT, ROOT, TOP_ROW, comm, &status);
			MPI_Send(&oldBacteriaGrid[rowsPerProcessor][1], colsPerProcessor, MPI_FLOAT, ROOT, BOT_ROW, comm); 
		}
		printf("Processor %d printing\n", procRank);
		printBacteriaMatrix(oldBacteriaGrid, rowsPerProcessor, colsPerProcessor);
		printf("\n");
		/* Diffuse nutrients in oldNutrientGrid */
		applyDiffusion(oldNutrientGrid, newNutrientGrid, diffusionRate, rowsPerProcessor, colsPerProcessor); /* stored in newNutrientGrid */
		/* Grow the bacteria from the oldBacteriaGrid using the newNutrientGrid */
		parallel2Grow(oldBacteriaGrid, newBacteriaGrid, newNutrientGrid, p, rowsPerProcessor, colsPerProcessor, procRank); /* stored in newBacteriaGrid */
		/* Bacteria on newBacterGrid have consumed nutrients in newNutrientGrid, so subtract from nutrients */
		consume(newNutrientGrid, oldNutrientGrid, newBacteriaGrid, rows, cols); /* stored in oldNutrientGrid */
	}
	printf("Processor %d done\n", procRank);
	MPI_Finalize();
	return 0;
}

float randomReal(void) {
	return sprng();
}

int random_int(int UpperBound) {
	float random_0_1;
	random_0_1 = (float) sprng() / RAND_MAX;
	return ((int)(UpperBound * random_0_1));
}

void printNutrientMatrix(float **matrix, int rows, int cols) {
	/* takes into account the ghost rows and cols */
	int i, j;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			printf("%3.2f ", matrix[i][j]);
		}
		printf("\n");
	}
}

void printBacteriaMatrix(int **matrix, int rows, int cols) {
	/* takes into account the ghost rows and cols */
	int i, j;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			printf("%d ", matrix[i][j]);
		}
		printf("\n");
	}
}

void initNutrientGrid(int rows, int cols, float **nutrientGrid) {
	int i, j;
	int rowsb = rows + 2;
	int colsb = cols + 2;
	// First and last boundary columns
	for (i = 0; i < rowsb; i++) {
		nutrientGrid[i][0] = 0;
		nutrientGrid[i][colsb - 1] = MAX_NUTRIENT;
	}
	// The center lattice
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			nutrientGrid[i][j] = MAX_NUTRIENT;
		}
	}
}

float diffusion(float diffusionRate, float site, float n, float ne, float e, float se, float s, float sw, float w, float nw) {
	return (1 - (8 * diffusionRate)) * site + (diffusionRate * (n + ne + e + se + s + sw + w + nw));
}

void applyDiffusion(float **oldNutrientGrid, float **newNutrientGrid, float diffusionRate, int rows, int cols) {
	int i, j, in, is, je, jw;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			in = i + 1;
			is = i - 1;
			je = j + 1;
			jw = j - 1;			
			newNutrientGrid[i][j] = diffusion(diffusionRate, oldNutrientGrid[i][j], oldNutrientGrid[in][j],
					oldNutrientGrid[in][je], oldNutrientGrid[i][je], oldNutrientGrid[is][je], oldNutrientGrid[is][j],
					oldNutrientGrid[is][jw], oldNutrientGrid[i][jw], oldNutrientGrid[in][jw]);
			
		}
	}
}

void initBacteriaGrid(int rows, int cols, int **bacteriaGrid, float probInitBacteria) {
	int i, j;
	int rowsb = rows + 2;
	int colsb = cols + 2;
	/* First and last boundary columns */
	for (i = 0; i < rowsb; i++) {
		bacteriaGrid[i][0] = BORDER;
		bacteriaGrid[i][colsb - 1] = BORDER;
	}
	/* First column of lattice */
	for (i = 1; i <= rows; i++) {
		if (randomReal() < probInitBacteria) {
			bacteriaGrid[i][1] = BACTERIUM;
		} else {
			bacteriaGrid[i][1] = EMPTY;
		}
	}
	/* Everything else */
	for (i = 1; i <= rows; i++) {
		for (j = 2; j <= cols; j++) {
			bacteriaGrid[i][j] = EMPTY;
		}
	}
	
}

float sumNutritionBacteria(int **bacteriaGrid, float **nutrientGrid, int rows, int cols) {
	int i, j;
	float sum = 0.0;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			if (bacteriaGrid[i][j] == BACTERIUM) {
				sum += nutrientGrid[i][j];
			}
		}
	}
	return sum;
}

float probGrow(int **bacteriaGrid, float **nutrientGrid, float p, int rows, int cols) {
	float sumOfNutrients = sumNutritionBacteria(bacteriaGrid, nutrientGrid, rows, cols);
	return (sumOfNutrients > 0) ? p / sumOfNutrients : 0.0;
}

GridLocation parallel2PickNeighbor(int row, int col, int rows, int cols, int n, int e, int s, int w) {
	int i, count;
	count = 0;
	int lst[4] = {n, e, s, w};
	int pos[4] = {0, 0, 0, 0};
	for (i = 0; i < 4; i++) {
		if (lst[i] == EMPTY) {
			++count;
			pos[i] = i;
		}
	}
	GridLocation loc;
	if (count == 0) {
		loc.row = row;
		loc.col = col;
	} else {
		/*int rand = random_int(count);*/
		int rand = random_int(count);
		int index = pos[rand];
		if (index == 0) {
			if (row > 0) {
				loc.row = row - 1;
				loc.col = col;
				loc.otherProc = 0;
			} else { /* row is the top row, so north is last row */
				loc.row = rows;
				loc.col = col;
				loc.otherProc = 1;
			}
		} else if (index == 1) {
			loc.row = row;
			loc.col = col + 1;
			loc.otherProc = 0;
		} else if (index == 2) {
			if (row <= rows) {
				loc.row = row + 1;
				loc.col = col;
				loc.otherProc = 0;
			} else { /* row is the bottom row, so south is top row */
				loc.row = 1;
				loc.col = col;
				loc.otherProc = 1;
			}
		} else { /* west */
			loc.row = row;
			loc.col = col - 1;
			loc.otherProc = 0;
		}
	}
	return loc;
}

void parallel2Grow(int **oldBacteriaGrid, int **newBacteriaGrid, float **nutrientGrid, float p, int rows, int cols, int procRank) {
	int i, j, in, is, je, jw, changed, index;
	float prob = probGrow(oldBacteriaGrid, nutrientGrid, p, rows, cols);
	int len = (2 * cols * 2) + 1; /* total possible size 2 columns and 1 for changed at beginning */
	int *locs = malloc(sizeof(int) * len); /* store any updated columns */ 
	int *locr = malloc(sizeof(int) * len);
	changed = 0;
	index = 1;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			if (oldBacteriaGrid[i][j] == BACTERIUM) {
				if (nutrientGrid[i][j] <= 0) {
					newBacteriaGrid[i][j] = DEAD;
				} else { /* try to divide */
					if (randomReal() < prob * nutrientGrid[i][j]) {
						in = i + 1;
						is = i - 1;
						je = j + 1;
						jw = j - 1;	
						GridLocation loc = parallel2PickNeighbor(i, j, rows, cols, oldBacteriaGrid[in][j], oldBacteriaGrid[i][je],
								oldBacteriaGrid[is][j], oldBacteriaGrid[i][jw]);
						if (loc.otherProc) {
							changed++;
							locs[index++] = loc.row;
							locs[index++] = loc.col;
						} else {
							newBacteriaGrid[loc.row][loc.col] = BACTERIUM;
							oldBacteriaGrid[loc.row][loc.col] = TAKEN;
						}
						newBacteriaGrid[i][j] = BACTERIUM;
					} else {
						newBacteriaGrid[i][j] = BACTERIUM;
					}
				}
				
			} else {
				if (oldBacteriaGrid[i][j] != TAKEN)
					newBacteriaGrid[i][j] = oldBacteriaGrid[i][j];
			}
		}
	}
	locs[0] = changed;
	MPI_Comm comm = MPI_COMM_WORLD;
	MPI_Status status;
	if (procRank == ROOT) {
		MPI_Send(locs, (changed * 2) + 1, MPI_INT, 1, CHANGED_ROOT, comm);
		MPI_Recv(locr, len, MPI_INT, 1, CHANGED_OTHER, comm, &status); 
	} else {
		MPI_Recv(locr, len, MPI_INT, ROOT, CHANGED_ROOT, comm, &status);
		MPI_Send(locs, (changed * 2) + 1, MPI_INT, ROOT, CHANGED_OTHER, comm);
	}
	processParallelGrowingCells(newBacteriaGrid, locs);
}

void processParallelGrowingCells(int **bacteriaGrid, int *locations) {
	int numGrew = locations[0];
	int i, row, col;
	for (i = 1; i <= numGrew * 2; i += 2) {
		row = locations[i];
		col = locations[i + 1];
		bacteriaGrid[row][col] = BACTERIUM;
	}
}

void consume(float **oldNutrientGrid, float **newNutrientGrid, int **bacteriaGrid, int rows, int cols) {
	int i, j;
	for (i = 1; i <= rows; i++) {
		for (j = 1; j <= cols; j++) {
			if (bacteriaGrid[i][j] == BACTERIUM) {
				newNutrientGrid[i][j] = max(0.0, oldNutrientGrid[i][j] - CONSUMED);
			} else {
				newNutrientGrid[i][j] = oldNutrientGrid[i][j];
			}
		}
	}
}
